"""Module for operations to fetch SPIM histology specimen procedures"""

import logging
from typing import List, Optional

from pydantic import BaseModel

from aind_slims_api import SlimsClient
from aind_slims_api.exceptions import SlimsRecordNotFound
from aind_slims_api.models.experiment_run_step import (
    SlimsExperimentRunStep,
    SlimsExperimentRunStepContent,
    SlimsExperimentTemplate,
    SlimsProtocolRunStep,
    SlimsWashRunStep,
)
from aind_slims_api.models.histology import (
    SlimsProtocolSOP,
    SlimsReagentContent,
    SlimsReagentDetailsRdrc,
    SlimsSampleContent,
    SlimsSource,
)


class SlimsReagent(BaseModel):
    """Pydantic model to store Reagent Info"""

    content: Optional[SlimsReagentContent] = None
    source: Optional[SlimsSource] = None
    details: Optional[SlimsReagentDetailsRdrc] = None


class SlimsWash(BaseModel):
    """Pydantic model to store Wash Info"""

    wash_step: Optional[SlimsWashRunStep] = None
    reagents: List[SlimsReagent] = []


class SPIMHistologyExpBlock(BaseModel):
    """Pydantic model to store Specimen Procedure Info"""

    protocol: Optional[SlimsProtocolSOP] = None
    washes: Optional[List[SlimsWash]] = None
    experiment_template: Optional[SlimsExperimentTemplate] = None


def fetch_reagents(client: SlimsClient, reagent_pks: List[int]) -> List[SlimsReagent]:
    """Fetches SlimsReagent objects for given reagent primary keys."""
    reagents = []
    for pk in reagent_pks:
        reagent_content = client.fetch_model(SlimsReagentContent, pk=pk)
        details = (
            client.fetch_model(
                SlimsReagentDetailsRdrc, pk=reagent_content.reagent_ref_pk
            )
            if reagent_content and reagent_content.reagent_ref_pk
            else None
        )
        source = (
            client.fetch_model(SlimsSource, pk=details.manufacturer_pk)
            if details and details.manufacturer_pk
            else None
        )
        reagents.append(
            SlimsReagent(
                content=reagent_content,
                details=details,
                source=source,
            )
        )
    return reagents


def fetch_washes(client: SlimsClient, experimentrun_pk: int) -> List[SlimsWash]:
    """Fetches washes for a given experimentrun_pk."""
    washes = []
    wash_run_steps = client.fetch_models(
        SlimsWashRunStep, experimentrun_pk=experimentrun_pk
    )
    for wash_step in wash_run_steps:
        reagents = (
            fetch_reagents(client, wash_step.reagent_pk) if wash_step.reagent_pk else []
        )
        washes.append(SlimsWash(wash_step=wash_step, reagents=reagents))

    return washes


def fetch_histology_procedures(
    client: SlimsClient, specimen_id: str
) -> List[SPIMHistologyExpBlock]:
    """
    Fetch and process all spim histology run steps for a given specimen id.
    Retrieves all SPIM histology steps associated with the provided specimen
    and returns a list of SPIMHistologyExpBlock objects.

    Parameters
    ----------
    client : SlimsClient
        An instance of SlimsClient used to connect to the SLIMS API.
    specimen_id : str
        The ID of the specimen for which to fetch histology data.

    Returns
    -------

    Example
    -------
    >>> from aind_slims_api import SlimsClient
    >>> client = SlimsClient()
    >>> specimen_procedures = fetch_histology_procedures(client, "000000")
    """
    specimen_procedures = []
    sample = client.fetch_model(SlimsSampleContent, mouse_barcode=specimen_id)

    content_runs = client.fetch_models(
        SlimsExperimentRunStepContent, mouse_pk=sample.pk
    )

    for content_run in content_runs:
        try:
            # retrieves content step to find experimentrun_pk
            content_run_step = client.fetch_model(
                SlimsExperimentRunStep, pk=content_run.runstep_pk
            )
            experiment_template = client.fetch_model(
                SlimsExperimentTemplate, pk=content_run_step.experiment_template_pk
            )
            protocol_run_step = client.fetch_model(
                SlimsProtocolRunStep, experimentrun_pk=content_run_step.experimentrun_pk
            )
            protocol_sop = None
            if protocol_run_step.protocol_pk:
                protocol_sop = client.fetch_model(
                    SlimsProtocolSOP, pk=protocol_run_step.protocol_pk
                )
            washes = fetch_washes(
                client, experimentrun_pk=content_run_step.experimentrun_pk
            )
            specimen_procedures.append(
                SPIMHistologyExpBlock(
                    protocol=protocol_sop,  # contains protocol link, name
                    experiment_template=experiment_template,
                    washes=washes,
                )
            )
        except SlimsRecordNotFound as e:
            logging.warning(str(e))
            continue

    return specimen_procedures



"""Module for operations to fetch SPIM histology specimen procedures"""

import logging
from aind_slims_api import SlimsClient
from aind_slims_api.exceptions import SlimsRecordNotFound
from aind_slims_api.models.experiment_run_step import (
    SlimsExperimentRunStep,
    SlimsExperimentRunStepContent,
    SlimsProtocolRunStep,
    SlimsSPIMImagingRunStep,
)
from typing import Dict, List
from aind_slims_api.models import (
    SlimsInstrumentRdrc,
    SlimsUser,
    SlimsProtocolSOP,
    SlimsSampleContent,
)
from aind_slims_api.models.imaging import (
    SlimsImagingMetadataResult,
    SlimsSPIMBrainOrientationRdrc,
)


def fetch_imaging_metadata(client: SlimsClient, subject_id: str) -> List[Dict]:
    """
    Fetch and process all spim imaging run steps for a given subject id.
    Retrieves all SPIM imaging steps associated with the provided specimen
    and returns a list of dictionaries.

    Parameters
    ----------
    client : SlimsClient
        An instance of SlimsClient used to connect to the SLIMS API.
    specimen_id : str
        The ID of the specimen for which to fetch histology data.
    Returns
    -------
    imaging_metadata : List[Dict]
    Example
    -------
    >>> from aind_slims_api import SlimsClient
    >>> client = SlimsClient()
    >>> imaging = fetch_imaging_metadata(client, "000000")
    """
    imaging_metadata = []
    sample = client.fetch_model(SlimsSampleContent, mouse_barcode=subject_id)

    content_runs = client.fetch_models(
        SlimsExperimentRunStepContent, mouse_pk=sample.pk
    )
    for content_run in content_runs:
        try:
            # retrieves content step to find experimentrun_pk
            content_run_step = client.fetch_model(
                SlimsExperimentRunStep, pk=content_run.runstep_pk
            )
            protocol_run_step = client.fetch_model(
                SlimsProtocolRunStep, experimentrun_pk=content_run_step.experimentrun_pk
            )
            protocol_sop = None
            if protocol_run_step.protocol_pk:
                protocol_sop = client.fetch_model(
                    SlimsProtocolSOP, pk=protocol_run_step.protocol_pk
                )
            imaging_steps = client.fetch_models(
                SlimsSPIMImagingRunStep,
                experimentrun_pk=content_run_step.experimentrun_pk,
            )
            for step in imaging_steps:
                imaging_results = client.fetch_models(
                    SlimsImagingMetadataResult,
                    experiment_run_step_pk=step.pk,
                    content_pk=sample.pk,
                )
                for imaging_result in imaging_results:
                    instrument = client.fetch_models(
                        SlimsInstrumentRdrc, pk=imaging_result.instrument_json_pk
                    )
                    surgeon = client.fetch_models(
                        SlimsUser, pk=imaging_result.surgeon_pk
                    )
                    brain_orientation = client.fetch_models(
                        SlimsSPIMBrainOrientationRdrc,
                        pk=imaging_result.brain_orientation_pk,
                    )
                    imaging_metadata.append(
                        {
                            "specimen_id": sample.barcode,
                            "protocol": protocol_sop,
                            "imaging_metadata": imaging_result,
                            "instrument": instrument[0].name if instrument else None,
                            "surgeon": surgeon[0].full_name if surgeon else None,
                            "brain_orientation": brain_orientation,
                        }
                    )
        except SlimsRecordNotFound as e:
            logging.warning(str(e))
            continue

    return imaging_metadata



"""
Module defining operations to build EcephysSession.
"""

import logging
from typing import List, Optional

from pydantic import BaseModel

from aind_slims_api import SlimsClient
from aind_slims_api.exceptions import SlimsRecordNotFound
from aind_slims_api.models import SlimsInstrumentRdrc, SlimsMouseContent
from aind_slims_api.models.ecephys_session import (
    SlimsBrainStructureRdrc,
    SlimsDomeModuleRdrc,
    SlimsMouseSessionResult,
    SlimsRewardDeliveryRdrc,
    SlimsRewardSpoutsRdrc,
    SlimsStimulusEpochsResult,
    SlimsStreamsResult,
)
from aind_slims_api.models.experiment_run_step import (
    SlimsExperimentRunStep,
    SlimsExperimentRunStepContent,
    SlimsGroupOfSessionsRunStep,
    SlimsMouseSessionRunStep,
)

logger = logging.getLogger(__name__)


class SlimsRewardDeliveryInfo(BaseModel):
    """Pydantic Model to store Reward Delivery Info"""

    reward_delivery: Optional[SlimsRewardDeliveryRdrc] = None
    reward_spouts: Optional[SlimsRewardSpoutsRdrc] = None


class SlimsStreamModule(SlimsDomeModuleRdrc):
    """DomeModule Wrapper to add linked brain structure models"""

    primary_targeted_structure: Optional[SlimsBrainStructureRdrc] = None
    secondary_targeted_structures: Optional[list[SlimsBrainStructureRdrc]] = []


class SlimsStream(SlimsStreamsResult):
    """Streams wrapper to add linked stream modules"""

    stream_modules: Optional[List[SlimsStreamModule]]


class EcephysSession(BaseModel):
    """
    Pydantic model encapsulating all session-related responses.
    """

    session_group: SlimsExperimentRunStep
    session_instrument: Optional[SlimsInstrumentRdrc] = None
    session_result: Optional[SlimsMouseSessionResult] = None
    streams: Optional[List[SlimsStream]] = []
    reward_delivery: Optional[SlimsRewardDeliveryInfo] = None
    stimulus_epochs: Optional[List[SlimsStimulusEpochsResult]] = []


class EcephysSessionBuilder:
    """Class to build EcephysSession objects from session run steps."""

    def __init__(self, client: SlimsClient):
        """Initialize Session Builder"""
        self.client = client

    def fetch_stream_modules(
        self, stream_modules_pk: list[int]
    ) -> List[SlimsStreamModule]:
        """Fetches stream modules and processes structure names."""
        stream_modules = (
            [
                self.client.fetch_model(SlimsDomeModuleRdrc, pk=pk)
                for pk in stream_modules_pk
            ]
            if stream_modules_pk
            else []
        )

        complete_stream_modules = []
        for stream_module in stream_modules:
            primary_structure, secondary_structures = None, []

            if stream_module.primary_targeted_structure_pk:
                primary_structure = self.client.fetch_model(
                    SlimsBrainStructureRdrc,
                    pk=stream_module.primary_targeted_structure_pk,
                )
                if stream_module.secondary_targeted_structures_pk:
                    secondary_structures = [
                        self.client.fetch_model(SlimsBrainStructureRdrc, pk=pk)
                        for pk in stream_module.secondary_targeted_structures_pk
                    ]

            stream_module_model = SlimsStreamModule(
                **stream_module.model_dump(),
                primary_targeted_structure=primary_structure,
                secondary_targeted_structures=secondary_structures,
            )
            complete_stream_modules.append(stream_module_model)
        return complete_stream_modules

    def fetch_streams(self, session_pk: int) -> List[SlimsStream]:
        """Fetches and completes stream information with modules."""
        streams = self.client.fetch_models(
            SlimsStreamsResult, mouse_session_pk=session_pk
        )
        complete_streams = [
            SlimsStream(
                **stream.model_dump(),
                stream_modules=(
                    self.fetch_stream_modules(stream.stream_modules_pk)
                    if stream.stream_modules_pk
                    else []
                ),
            )
            for stream in streams
        ]
        return complete_streams

    def fetch_reward_data(self, reward_delivery_pk: int) -> SlimsRewardDeliveryInfo:
        """Fetches reward delivery and spouts data."""
        reward_delivery = self.client.fetch_model(
            SlimsRewardDeliveryRdrc, pk=reward_delivery_pk
        )
        reward_spouts = (
            self.client.fetch_model(
                SlimsRewardSpoutsRdrc, pk=reward_delivery.reward_spouts_pk
            )
            if reward_delivery and reward_delivery.reward_spouts_pk
            else None
        )
        return SlimsRewardDeliveryInfo(
            reward_delivery=reward_delivery, reward_spouts=reward_spouts
        )

    def _process_single_step(self, group_run_step, session_run_step) -> EcephysSession:
        """Process a single session run step into an EcephysSession."""
        session = self.client.fetch_model(
            SlimsMouseSessionResult, experiment_run_step_pk=session_run_step.pk
        )
        session_instrument = self.client.fetch_model(
            SlimsInstrumentRdrc, pk=group_run_step.instrument_pk
        )
        stimulus_epochs = self.client.fetch_models(
            SlimsStimulusEpochsResult, mouse_session_pk=session.pk
        )

        streams = self.fetch_streams(session.pk)
        reward_delivery = (
            self.fetch_reward_data(session.reward_delivery_pk)
            if session.reward_delivery_pk
            else None
        )

        return EcephysSession(
            session_group=group_run_step,
            session_instrument=session_instrument or None,
            session_result=session,
            streams=streams or None,
            reward_delivery=reward_delivery,
            stimulus_epochs=stimulus_epochs or [],
        )

    def process_session_steps(
        self,
        group_run_step: SlimsGroupOfSessionsRunStep,
        session_run_steps: List[SlimsMouseSessionRunStep],
    ) -> List[EcephysSession]:
        """
        Processes all session run steps into EcephysSession objects.
        Parameters
        ----------
        group_run_step : SlimsGroupOfSessionsRunStep
            The group run step containing session metadata and run information.
        session_run_steps : List[SlimsMouseSessionRunStep]
            A list of individual session run steps to be processed and encapsulated.

        Returns
        -------
        List[EcephysSession]
            A list of EcephysSession objects containing the processed session data.
        """
        return [
            self._process_single_step(group_run_step, step)
            for step in session_run_steps
        ]


def fetch_ecephys_sessions(
    client: SlimsClient, subject_id: str
) -> List[EcephysSession]:
    """
    Fetch and process all electrophysiology (ecephys) run steps for a given subject.
    Retrieves all electrophysiology sessions associated with the provided subject ID
    and returns a list of EcephysSession objects.

    Parameters
    ----------
    client : SlimsClient
        An instance of SlimsClient used to connect to the SLIMS API.
    subject_id : str
        The ID of the subject (mouse) for which to fetch electrophysiology session data.

    Returns
    -------
    List[EcephysSession]
        A list of EcephysSession objects containing data for each run step.

    Example
    -------
    >>> from aind_slims_api import SlimsClient
    >>> client = SlimsClient()
    >>> sessions = fetch_ecephys_sessions(client=client, subject_id="000000")
    """
    ecephys_sessions_list = []
    mouse = client.fetch_model(SlimsMouseContent, barcode=subject_id)
    content_runs = client.fetch_models(SlimsExperimentRunStepContent, mouse_pk=mouse.pk)

    for content_run in content_runs:
        try:
            # retrieves content step to find experimentrun_pk
            content_run_step = client.fetch_model(
                SlimsExperimentRunStep, pk=content_run.runstep_pk
            )

            # retrieve group and mouse sessions in the experiment run
            group_run_step = client.fetch_model(
                SlimsGroupOfSessionsRunStep,
                experimentrun_pk=content_run_step.experimentrun_pk,
            )
            session_run_steps = client.fetch_models(
                SlimsMouseSessionRunStep,
                experimentrun_pk=content_run_step.experimentrun_pk,
            )
            if group_run_step and session_run_steps:
                esb = EcephysSessionBuilder(client=client)
                ecephys_sessions = esb.process_session_steps(
                    group_run_step=group_run_step,
                    session_run_steps=session_run_steps,
                )
                ecephys_sessions_list.extend(ecephys_sessions)

        except SlimsRecordNotFound as e:
            logging.warning(str(e))
            continue

    return ecephys_sessions_list



"""Contains models for a histology result stored in SLIMS."""

from datetime import datetime
from typing import Any, ClassVar, Optional

from pydantic import Field

from aind_slims_api.models.base import SlimsBaseModel


class SlimsSampleContent(SlimsBaseModel):
    """Model for an instance of the Whole Brain ContentType

    Properties
    ----------
    barcode: str, barcode of the mouse, filterable

    Examples
    --------
    >>> from aind_slims_api.core import SlimsClient
    >>> client = SlimsClient()
    >>> mouse = client.fetch_model(SlimsSampleContent, barcode="00000000")
    """

    barcode: str = Field(
        ...,
        serialization_alias="cntn_barCode",
        validation_alias="cntn_barCode",
    )
    mouse_barcode: Optional[str] = Field(
        None,
        serialization_alias="cntn_cf_parentBarcode",
        validation_alias="cntn_cf_parentBarcode",
    )
    pk: Optional[int] = Field(
        default=None,
        serialization_alias="cntn_pk",
        validation_alias="cntn_pk",
    )
    content_type: Optional[int] = Field(
        default=None,
        serialization_alias="cntn_fk_contentType",
        validation_alias="cntn_fk_contentType",
    )
    created_on: Optional[datetime] = Field(
        None,
        serialization_alias="cntn_createdOn",
        validation_alias="cntn_createdOn",
    )
    category_name: Optional[str] = Field(
        None, serialization_alias="category_name", validation_alias="category_name"
    )
    collection_date: Optional[datetime] = Field(
        None,
        serialization_alias="cntn_collectionDate",
        validation_alias="cntn_collectionDate",
    )
    _slims_table = "Content"
    _base_fetch_filters: ClassVar[dict[str, str]] = {
        "cntp_name": "Whole Brain",
    }


class SlimsReagentContent(SlimsBaseModel):
    """Model for a Reagent Content"""

    barcode: str = Field(
        ...,
        serialization_alias="cntn_barCode",
        validation_alias="cntn_barCode",
    )
    pk: Optional[int] = Field(
        default=None,
        serialization_alias="cntn_pk",
        validation_alias="cntn_pk",
    )
    content_type: Optional[int] = Field(
        default=None,
        serialization_alias="cntn_fk_contentType",
        validation_alias="cntn_fk_contentType",
    )
    created_on: Optional[datetime] = Field(
        None,
        serialization_alias="cntn_createdOn",
        validation_alias="cntn_createdOn",
    )
    lot_number: Optional[str] = Field(
        None,
        serialization_alias="cntn_cf_lotNumber",
        validation_alias="cntn_cf_lotNumber",
    )
    reagent_ref_pk: Optional[int] = Field(
        None,
        serialization_alias="cntn_cf_fk_reagentCatalogNumber",
        validation_alias="cntn_cf_fk_reagentCatalogNumber",
    )
    source_pk: Optional[int] = Field(
        None, serialization_alias="cntn_fk_source", validation_alias="cntn_fk_source"
    )
    _slims_table = "Content"


class SlimsSource(SlimsBaseModel):
    """Model for a Source"""

    pk: Optional[int] = Field(
        default=None,
        serialization_alias="sorc_pk",
        validation_alias="sorc_pk",
    )
    name: Optional[str] = Field(
        None, serialization_alias="sorc_name", validation_alias="sorc_name"
    )
    created_on: Optional[datetime] = Field(
        default=None,
        serialization_alias="sorc_createdOn",
        validation_alias="sorc_createdOn",
    )
    _slims_table = "Source"


class SlimsReagentDetailsRdrc(SlimsBaseModel):
    """Model for a Reagent Catalog Reference Data Record"""

    pk: Optional[int] = Field(
        None, serialization_alias="rdrc_pk", validation_alias="rdrc_pk"
    )
    name: Optional[str] = Field(
        None, serialization_alias="rdrc_name", validation_alias="rdrc_name"
    )
    manufacturer_pk: Optional[int] = Field(
        None,
        serialization_alias="rdrc_cf_fk_manufacturer",
        validation_alias="rdrc_cf_fk_manufacturer",
    )
    created_on: Optional[datetime] = Field(
        None,
        serialization_alias="rdrc_createdOn",
        validation_alias="rdrc_createdOn",
    )
    _slims_table = "ReferenceDataRecord"
    _base_fetch_filters: ClassVar[dict[str, Any]] = {
        "rdty_name": [
            "External reagent details",
            "Reagent Details",
            "Internally Produced Reagent Details",
        ]
    }


class SlimsProtocolSOP(SlimsBaseModel):
    """Model for Protocols SOP"""

    pk: Optional[int] = Field(
        default=None,
        serialization_alias="stop_pk",
        validation_alias="stop_pk",
    )
    name: Optional[str] = Field(
        None, serialization_alias="stop_name", validation_alias="stop_name"
    )
    link: Optional[str] = Field(
        None, serialization_alias="stop_link", validation_alias="stop_link"
    )
    created_on: Optional[datetime] = Field(
        None,
        serialization_alias="stop_createdOn",
        validation_alias="stop_createdOn",
    )
    _slims_table = "SOP"



"""Contains models for Imaging in SLIMS."""

from datetime import datetime
from typing import ClassVar, Optional
from pydantic import Field
from aind_slims_api.models.base import SlimsBaseModel


class SlimsImagingMetadataResult(SlimsBaseModel):
    """Model for a SLIMS Imaging Metadata"""

    pk: Optional[int] = Field(
        default=None, serialization_alias="rslt_pk", validation_alias="rslt_pk"
    )
    created_on: Optional[datetime] = Field(
        default=None,
        serialization_alias="rslt_createdOn",
        validation_alias="rslt_createdOn",
    )
    modified_by: Optional[str] = Field(
        default=None,
        serialization_alias="rslt_modifiedBy",
        validation_alias="rslt_modifiedBy",
    )
    content_pk: Optional[int] = Field(
        default=None,
        serialization_alias="rslt_fk_content",
        validation_alias="rslt_fk_content",
    )
    date_performed: Optional[datetime] = Field(
        default=None,
        serialization_alias="rslt_cf_datePerformed",
        validation_alias="rslt_cf_datePerformed",
    )
    sample_refractive_index: Optional[float] = Field(
        default=None,
        serialization_alias="rslt_cf_sampleRefractiveIndex1",
        validation_alias="rslt_cf_sampleRefractiveIndex1",
    )
    sample_immersion_medium: Optional[str] = Field(
        default=None,
        serialization_alias="rslt_cf_sampleImmersionMedium",
        validation_alias="rslt_cf_sampleImmersionMedium",
    )
    chamber_refractive_index: Optional[float] = Field(
        default=None,
        serialization_alias="rslt_cf_chamberRefractiveIndex",
        validation_alias="rslt_cf_chamberRefractiveIndex",
    )
    chamber_immersion_medium: Optional[str] = Field(
        default=None,
        serialization_alias="rslt_cf_chamberImmersionMedium",
        validation_alias="rslt_cf_chamberImmersionMedium",
    )
    notes: Optional[str] = Field(
        default=None,
        serialization_alias="rslt_comments",
        validation_alias="rslt_comments",
    )
    experiment_run_step_pk: Optional[int] = Field(
        default=None,
        serialization_alias="rslt_fk_experimentRunStep",
        validation_alias="rslt_fk_experimentRunStep",
    )
    surgeon_pk: Optional[int] = Field(
        default=None,
        serialization_alias="rslt_cf_fk_surgeon",
        validation_alias="rslt_cf_fk_surgeon",
    )
    brain_orientation_pk: Optional[int] = Field(
        default=None,
        serialization_alias="rslt_cf_fk_spimBrainOrientation",
        validation_alias="rslt_cf_fk_spimBrainOrientation",
    )
    instrument_json_pk: Optional[int] = Field(
        default=None,
        serialization_alias="rslt_cf_fk_instrumentJson",
        validation_alias="rslt_cf_fk_instrumentJson",
    )
    _slims_table = "Result"


class SlimsSPIMBrainOrientationRdrc(SlimsBaseModel):
    """Model for a SLIMS Brain Orientation"""

    pk: Optional[int] = Field(
        default=None, serialization_alias="rdrc_pk", validation_alias="rdrc_pk"
    )
    created_on: Optional[datetime] = Field(
        default=None,
        serialization_alias="rdrc_createdOn",
        validation_alias="rdrc_createdOn",
    )
    modified_by: Optional[str] = Field(
        default=None,
        serialization_alias="rdrc_modifiedBy",
        validation_alias="rdrc_modifiedBy",
    )
    name: Optional[str] = Field(
        default=None,
        serialization_alias="rdrc_name",
        validation_alias="rdrc_name",
    )
    z_direction: Optional[str] = Field(
        default=None,
        serialization_alias="rdrc_cf_washType",
        validation_alias="rdrc_cf_washType",
    )
    y_direction: Optional[str] = Field(
        default=None,
        validation_alias="rdrc_cf_additionalWashReagent",
        serialization_alias="rdrc_cf_additionalWashReagent",
    )
    x_direction: Optional[str] = Field(
        default=None,
        serialization_alias="rdrc_cf_hostSpecies",
        validation_alias="rdrc_cf_hostSpecies",
    )
    _slims_table = "ReferenceDataRecord"
    _base_fetch_filters: ClassVar[dict[str, str]] = {
        "rdty_name": "SPIM Brain Orientation",
    }



"""Tests methods in histology procedures operation"""

import os
import unittest
from pathlib import Path
from unittest.mock import patch

from aind_slims_api.exceptions import SlimsRecordNotFound
from aind_slims_api.models.experiment_run_step import (
    SlimsExperimentRunStep,
    SlimsExperimentRunStepContent,
    SlimsExperimentTemplate,
    SlimsProtocolRunStep,
    SlimsWashRunStep,
)
from aind_slims_api.models.histology import (
    SlimsProtocolSOP,
    SlimsReagentContent,
    SlimsReagentDetailsRdrc,
    SlimsSampleContent,
    SlimsSource,
)
from aind_slims_api.operations.histology_procedures import (
    SlimsReagent,
    SlimsWash,
    fetch_histology_procedures,
    fetch_reagents,
    fetch_washes,
)

RESOURCES_DIR = Path(os.path.dirname(os.path.realpath(__file__))) / ".." / "resources"


class TestHistologyProcedures(unittest.TestCase):
    """Test class for SlimsHistologyProcedures operation."""

    @patch("aind_slims_api.operations.histology_procedures.SlimsClient")
    def setUp(cls, mock_client):
        """setup test class"""
        cls.client = mock_client()
        cls.example_reagent_content = SlimsReagentContent(
            pk=123,
            source_pk=456,
            lot_number="EI60",
            barcode="0000000",
            reagent_ref_pk=10,
        )
        cls.example_reagent_details = SlimsReagentDetailsRdrc(
            pk=10,
            manufacturer_pk=456,
            name="Shield On",
        )
        cls.example_reagent_source = SlimsSource(
            pk=456,
            name="AA Opto Electronics",
        )
        cls.example_wash_run_step = SlimsWashRunStep(
            reagent_pk=[123],
            experimentrun_pk=789,
            wash_name="Wash 1",
            spim_wash_type="Passive Delipidation",
        )

    @patch("aind_slims_api.operations.histology_procedures.SlimsReagent")
    @patch("aind_slims_api.operations.histology_procedures.SlimsWash")
    def test_fetch_washes(self, mock_slims_wash, mock_slims_reagent):
        """Tests washes are fetched successfully."""
        self.client.fetch_models.side_effect = lambda model, **kwargs: (
            [self.example_wash_run_step] if model == SlimsWashRunStep else []
        )
        self.client.fetch_model.side_effect = lambda model, **kwargs: {
            SlimsReagentContent: self.example_reagent_content,
            SlimsReagentDetailsRdrc: self.example_reagent_details,
            SlimsSource: self.example_reagent_source,
        }.get(model)

        washes = fetch_washes(self.client, experimentrun_pk=789)

        self.client.fetch_models.assert_any_call(SlimsWashRunStep, experimentrun_pk=789)
        self.client.fetch_model.assert_any_call(SlimsReagentContent, pk=123)
        self.client.fetch_model.assert_any_call(SlimsReagentDetailsRdrc, pk=10)
        self.client.fetch_model.assert_any_call(SlimsSource, pk=456)
        mock_slims_reagent.assert_called_with(
            content=self.example_reagent_content,
            details=self.example_reagent_details,
            source=self.example_reagent_source,
        )
        mock_slims_wash.assert_called_with(
            wash_step=self.example_wash_run_step,
            reagents=[mock_slims_reagent.return_value],
        )
        self.assertEqual(len(washes), 1)

    @patch("aind_slims_api.operations.histology_procedures.SlimsReagent")
    def test_fetch_reagents(self, mock_slims_reagent):
        """Tests reagents are fetched successfully."""
        self.client.fetch_model.side_effect = lambda model, **kwargs: {
            SlimsReagentContent: self.example_reagent_content,
            SlimsReagentDetailsRdrc: self.example_reagent_details,
            SlimsSource: self.example_reagent_source,
        }.get(model)

        reagents = fetch_reagents(self.client, reagent_pks=[123])

        self.client.fetch_model.assert_any_call(SlimsReagentContent, pk=123)
        self.client.fetch_model.assert_any_call(SlimsReagentDetailsRdrc, pk=10)
        self.client.fetch_model.assert_any_call(SlimsSource, pk=456)
        mock_slims_reagent.assert_called_with(
            content=self.example_reagent_content,
            details=self.example_reagent_details,
            source=self.example_reagent_source,
        )
        self.assertEqual(len(reagents), 1)
        self.assertEqual(reagents[0], mock_slims_reagent.return_value)

    @patch("aind_slims_api.operations.histology_procedures.fetch_washes")
    def test_fetch_histology_procedures(self, mock_fetch_washes):
        """Tests that specimen procedures are fetched as expected."""
        example_sample_content = SlimsSampleContent(
            pk=1, mouse_barcode="000000", barcode="000000"
        )
        example_run_step = SlimsExperimentRunStep(
            experiment_template_pk=1426, experimentrun_pk=789
        )
        example_experiment_template = SlimsExperimentTemplate(pk=1426, name="SmartSPIM")
        example_protocol_run_step = SlimsProtocolRunStep(protocol_pk=101)
        example_protocol_sop = SlimsProtocolSOP(pk=101, name="LifeCanvas Clearing")

        self.client.fetch_model.side_effect = lambda model, **kwargs: (
            example_sample_content
            if model == SlimsSampleContent
            else (
                example_run_step
                if model == SlimsExperimentRunStep
                else (
                    example_experiment_template
                    if model == SlimsExperimentTemplate
                    else (
                        example_protocol_run_step
                        if model == SlimsProtocolRunStep
                        else example_protocol_sop if model == SlimsProtocolSOP else None
                    )
                )
            )
        )

        self.client.fetch_models.side_effect = lambda model, **kwargs: (
            [SlimsExperimentRunStepContent(runstep_pk=i) for i in range(1, 7)]
            if model == SlimsExperimentRunStepContent
            else []
        )

        # Mock fetch_washes to return valid SlimsWash objects
        mock_fetch_washes.side_effect = lambda c, experimentrun_pk: [
            SlimsWash(
                wash_step=SlimsWashRunStep(
                    reagent_pk=[123],
                    experimentrun_pk=experimentrun_pk,
                    name=f"Wash {experimentrun_pk}",
                    spim_wash_type="Passive Delipidation",
                ),
                reagents=[
                    SlimsReagent(
                        content=self.example_reagent_content,
                        source=self.example_reagent_source,
                        details=self.example_reagent_details,
                    )
                ],
            )
            for _ in range(4)  # Simulate 4 washes
        ]
        result = fetch_histology_procedures(self.client, specimen_id="000000")

        self.assertEqual(len(result), 6)
        for result_block in result:
            # Check protocol and experiment template
            self.assertEqual(result_block.protocol.name, "LifeCanvas Clearing")
            self.assertEqual(result_block.experiment_template.pk, 1426)

            # Check if washes have been populated correctly
            self.assertEqual(len(result_block.washes), 4)
            for wash in result_block.washes:
                self.assertIsInstance(wash, SlimsWash)
                self.assertIsNotNone(wash.wash_step)
                self.assertEqual(wash.wash_step.spim_wash_type, "Passive Delipidation")
                self.assertGreater(len(wash.reagents), 0)
                for reagent in wash.reagents:
                    self.assertIsInstance(reagent, SlimsReagent)
                    self.assertIsInstance(reagent.content, SlimsReagentContent)
                    self.assertIsInstance(reagent.source, SlimsSource)
                    self.assertIsInstance(reagent.details, SlimsReagentDetailsRdrc)
        # Check attributes of first wash
        first_wash = result[0].washes[0]
        self.assertEqual(first_wash.wash_step.name, "Wash 789")
        self.assertEqual(first_wash.wash_step.experimentrun_pk, 789)
        self.assertEqual(first_wash.reagents[0].content.lot_number, "EI60")
        self.assertEqual(first_wash.reagents[0].source.name, "AA Opto Electronics")
        self.assertEqual(first_wash.reagents[0].details.name, "Shield On")

    def test_fetch_histology_procedures_handles_exception(self):
        """Tests that exception is handled as expected"""
        self.client.fetch_models.side_effect = [
            [SlimsExperimentRunStepContent(pk=1, runstep_pk=3, mouse_pk=67890)]
        ]
        self.client.fetch_model.side_effect = [
            SlimsSampleContent.model_construct(pk=67890),
            SlimsRecordNotFound("No record found for SlimsExperimentRunStep with pk=3"),
        ]

        with patch("logging.warning") as mock_log_warning:
            fetch_histology_procedures(client=self.client, specimen_id="67890")
            mock_log_warning.assert_called_with(
                "No record found for SlimsExperimentRunStep with pk=3"
            )


if __name__ == "__main__":
    unittest.main()



"""Tests methods in spim imaging operation"""

import unittest
from unittest.mock import patch
from aind_slims_api.operations.spim_imaging import fetch_imaging_metadata
from aind_slims_api.models.experiment_run_step import (
    SlimsExperimentRunStep,
    SlimsExperimentRunStepContent,
    SlimsExperimentTemplate,
    SlimsProtocolRunStep,
    SlimsSPIMImagingRunStep,
)
from aind_slims_api.models.imaging import (
    SlimsImagingMetadataResult,
    SlimsSPIMBrainOrientationRdrc,
)
from aind_slims_api.models import (
    SlimsInstrumentRdrc,
    SlimsUser,
    SlimsSampleContent,
    SlimsProtocolSOP,
)
from aind_slims_api.exceptions import SlimsRecordNotFound


class TestFetchImagingMetadata(unittest.TestCase):
    """Test class for fetch_imaging_metadata operation."""

    @patch("aind_slims_api.operations.spim_imaging.SlimsClient")
    def setUp(self, mock_client):
        """Setup test class"""
        self.client = mock_client()

        # Mock sample data
        self.example_sample_content = SlimsSampleContent(
            pk=1, mouse_barcode="000000", barcode="000000"
        )
        self.example_run_step = SlimsExperimentRunStep(
            experiment_template_pk=1426, experimentrun_pk=789
        )
        self.example_experiment_template = SlimsExperimentTemplate(
            pk=1426, name="SPIMImaging"
        )
        self.example_protocol_run_step = SlimsProtocolRunStep(protocol_pk=101)
        self.example_protocol_sop = SlimsProtocolSOP(pk=101, name="Some Protocol SOP")
        self.imaging_step = SlimsSPIMImagingRunStep(pk=6)
        self.imaging_result = SlimsImagingMetadataResult(
            pk=7, instrument_json_pk=8, surgeon_pk=9, brain_orientation_pk=10
        )
        self.instrument = SlimsInstrumentRdrc(pk=8, name="Instrument A")
        self.surgeon = SlimsUser(pk=9, full_name="Surgeon 1", username="surgeon1")
        self.brain_orientation = SlimsSPIMBrainOrientationRdrc(
            pk=10,
            name="Horizontal, Superior; AP",
            x_direction="Left to Right",
            y_direction="Anterior to Posterior",
            z_direction="Superior to Inferior",
        )

    def test_fetch_imaging_metadata_success(self):
        """ "Tests fetch imaging operation succeeds."""
        self.client.fetch_model.side_effect = lambda model, **kwargs: (
            self.example_sample_content
            if model == SlimsSampleContent
            else (
                self.example_run_step
                if model == SlimsExperimentRunStep
                else (
                    self.example_experiment_template
                    if model == SlimsExperimentTemplate
                    else (
                        self.example_protocol_run_step
                        if model == SlimsProtocolRunStep
                        else (
                            self.example_protocol_sop
                            if model == SlimsProtocolSOP
                            else None
                        )
                    )
                )
            )
        )
        self.client.fetch_models.side_effect = lambda model, **kwargs: (
            [SlimsExperimentRunStepContent(runstep_pk=123)]
            if model == SlimsExperimentRunStepContent
            else (
                [self.imaging_step]
                if model == SlimsSPIMImagingRunStep
                else (
                    [self.imaging_result]
                    if model == SlimsImagingMetadataResult
                    else (
                        [self.instrument]
                        if model == SlimsInstrumentRdrc
                        else (
                            [self.surgeon]
                            if model == SlimsUser
                            else (
                                [self.brain_orientation]
                                if model == SlimsSPIMBrainOrientationRdrc
                                else []
                            )
                        )
                    )
                )
            )
        )

        metadata = fetch_imaging_metadata(self.client, "000000")
        self.assertEqual(len(metadata), 1)
        self.assertEqual(metadata[0]["instrument"], "Instrument A")
        self.assertEqual(metadata[0]["surgeon"], "Surgeon 1")
        self.assertEqual(metadata[0]["brain_orientation"], [self.brain_orientation])

    def test_fetch_imaging_metadata_no_content(self):
        """Tests case when specimen has no content runs"""
        self.client.fetch_model.return_value = self.example_sample_content
        self.client.fetch_models.side_effect = lambda model, **kwargs: {
            SlimsExperimentRunStepContent: [],
        }.get(model, [])

        metadata = fetch_imaging_metadata(self.client, "000000")
        self.assertEqual(metadata, [])

    def test_fetch_imaging_metadata_missing_record(self):
        """Tests that exception is handled as expected"""
        self.client.fetch_models.side_effect = [
            [SlimsExperimentRunStepContent(pk=1, runstep_pk=3, mouse_pk=67890)]
        ]
        self.client.fetch_model.side_effect = [
            SlimsSampleContent.model_construct(pk=67890),
            SlimsRecordNotFound("No record found for SlimsExperimentRunStep with pk=3"),
        ]

        with patch("logging.warning") as mock_log_warning:
            fetch_imaging_metadata(client=self.client, subject_id="67890")
            mock_log_warning.assert_called_with(
                "No record found for SlimsExperimentRunStep with pk=3"
            )


if __name__ == "__main__":
    unittest.main()



"""Contents:

Utilities for creating pydantic models for SLIMS data:
    SlimsBaseModel - to be subclassed for SLIMS pydantic models
    UnitSpec - To be included in a type annotation of a Quantity field

SlimsClient - Basic wrapper around slims-python-api client with convenience
    methods and integration with SlimsBaseModel subtypes
"""

import base64
import logging
from copy import deepcopy
from functools import lru_cache
from typing import Any, Optional, Type, TypeVar

from pydantic import ValidationError
from requests import Response
from slims.criteria import Criterion, conjunction, equals, is_one_of
from slims.internal import Record as SlimsRecord
from slims.slims import Slims, _SlimsApiException

from aind_slims_api import config
from aind_slims_api.exceptions import SlimsRecordNotFound
from aind_slims_api.filters import resolve_filter_args
from aind_slims_api.models import SlimsAttachment
from aind_slims_api.models.base import SlimsBaseModel
from aind_slims_api.types import SLIMS_TABLES

logger = logging.getLogger(__name__)


SlimsBaseModelTypeVar = TypeVar("SlimsBaseModelTypeVar", bound=SlimsBaseModel)


class SlimsClient:
    """Wrapper around slims-python-api client with convenience methods"""

    db: Slims

    def __init__(self, url=None, username=None, password=None):
        """Create object and try to connect to database"""
        self.url = url or config.slims_url

        self.connect(
            self.url,
            username or config.slims_username,
            password or config.slims_password.get_secret_value(),
        )

    def connect(self, url: str, username: str, password: str):
        """Connect to the database"""
        self.db = Slims(
            "slims",
            url,
            username,
            password,
        )

    def fetch(
        self,
        table: SLIMS_TABLES,
        *args,
        sort: Optional[str | list[str]] = None,
        start: Optional[int] = None,
        end: Optional[int] = None,
        **kwargs,
    ) -> list[SlimsRecord]:
        """Fetch from the SLIMS database

        Args:
            table (str): SLIMS table to query
            sort (str | list[str], optional): Fields to sort by; e.g. date
            start (int, optional):  The first row to return
            end (int, optional): The last row to return
            *args (Slims.criteria.Criterion): Optional criteria to apply
            **kwargs (dict[str,str|list]): "field=value" filters.
                If value is a list, will apply "field IN value" criterion


        Returns:
            records (list[SlimsRecord] | None): Matching records, if any
        """
        criteria = conjunction()
        for arg in args:
            if isinstance(arg, Criterion):
                criteria.add(arg)

        for k, v in kwargs.items():
            if isinstance(v, list):  # Handle lists as "IN" criteria
                criteria.add(is_one_of(k, v))
            else:
                criteria.add(equals(k, v))

        if isinstance(sort, str):
            sort = [sort]

        try:
            records = self.db.fetch(
                table,
                criteria,
                sort=sort,
                start=start,
                end=end,
            )
        except _SlimsApiException as e:
            # TODO: Add better error handling
            #  Let's just raise error for the time being
            raise e

        return records

    @staticmethod
    def _validate_models(
        model_type: Type[SlimsBaseModelTypeVar], records: list[SlimsRecord]
    ) -> list[SlimsBaseModelTypeVar]:
        """Validate a list of SlimsBaseModel objects. Logs errors for records
        that fail pydantic validation."""
        validated = []
        for record in records:
            try:
                validated.append(model_type.model_validate(record))
            except ValidationError as e:
                logger.error(f"SLIMS data validation failed, {repr(e)}")
        return validated

    def fetch_models(
        self,
        model: Type[SlimsBaseModelTypeVar],
        *args: Criterion,
        sort: str | list[str] = [],
        start: Optional[int] = None,
        end: Optional[int] = None,
        **kwargs,
    ) -> list[SlimsBaseModelTypeVar]:
        """Fetch records from SLIMS and return them as SlimsBaseModel objects.

        Returns
        -------
        tuple:
            list:
                Validated SlimsBaseModel objects

        Notes
        -----
        - kwargs are mapped to field alias names and used as equality filters
         for the fetch.
        """
        resolved_kwargs = deepcopy(model._base_fetch_filters)
        logger.debug("Resolved kwargs: %s", resolved_kwargs)

        if isinstance(sort, str):
            sort = [sort]

        criteria, resolved_sort, start, end = resolve_filter_args(
            model,
            *args,
            sort=sort,
            start=start,
            end=end,
            **kwargs,
        )
        response = self.fetch(
            model._slims_table,
            *criteria,
            sort=resolved_sort,
            start=start,
            end=end,
            **resolved_kwargs,
        )
        return self._validate_models(model, response)

    def fetch_model(
        self,
        model: Type[SlimsBaseModelTypeVar],
        *args: Criterion,
        **kwargs,
    ) -> SlimsBaseModelTypeVar | None:
        """Fetch a single record from SLIMS and return it as a validated
         SlimsBaseModel object.

        Notes
        -----
        - kwargs are mapped to field alias values
        - sorts records on created_on in descending order and returns the first
        """
        records = self.fetch_models(
            model,
            *args,
            sort="-created_on",
            start=0,  # slims rows appear to be 0-indexed
            end=1,
            **kwargs,
        )
        if len(records) > 0:
            logger.debug(f"Found {len(records)} records for {model}.")
        if len(records) < 1:
            raise SlimsRecordNotFound(f"No record found for {model} with {args}.")
        return records[0]

    @staticmethod
    def _create_get_entities_body(
        *args: Criterion,
        sort: list[str] = [],
        start: Optional[int] = None,
        end: Optional[int] = None,
    ) -> dict[str, Any]:
        """Creates get entities body for SLIMS API request."""
        body: dict[str, Any] = {
            "sortBy": sort,
            "startRow": start,
            "endRow": end,
        }
        if args:
            criteria = conjunction()
            for arg in args:
                criteria.add(arg)
            body["criteria"] = criteria.to_dict()

        return body

    def fetch_attachments(
        self,
        record: SlimsBaseModel,
        *args: Criterion,
        sort: str | list[str] = [],
        start: Optional[int] = None,
        end: Optional[int] = None,
        **kwargs,
    ) -> list[SlimsAttachment]:
        """Fetch attachments for a given record.

        Notes
        -----
        - kwargs are mapped to field alias values
        """
        if isinstance(sort, str):
            sort = [sort]

        criteria, sort, start, end = resolve_filter_args(
            SlimsAttachment,
            *args,
            sort=sort,
            start=start,
            end=end,
            **kwargs,
        )
        return self._validate_models(
            SlimsAttachment,
            self.db.slims_api.get_entities(
                f"attachment/{record._slims_table}/{record.pk}",
                body=self._create_get_entities_body(
                    *criteria,
                    sort=sort,
                    start=start,
                    end=end,
                ),
            ),
        )

    def fetch_attachment(
        self,
        record: SlimsBaseModel,
        *args: Criterion,
        **kwargs,
    ) -> SlimsAttachment:
        """Fetch attachments for a given record.

        Notes
        -----
        - kwargs are mapped to field alias values
        - sorts records on created_on in descending order and returns the first
        """
        records = self.fetch_attachments(
            record,
            *args,
            sort="-created_on",
            start=0,  # slims rows appear to be 0-indexed
            end=1,
            **kwargs,
        )
        if len(records) > 0:
            logger.debug(f"Found {len(records)} records for {record}.")
        if len(records) < 1:
            raise SlimsRecordNotFound("No record found.")
        return records[0]

    def fetch_attachment_content(
        self,
        attachment: int | SlimsAttachment,
    ) -> Response:
        """Fetch attachment content for a given attachment.

        Parameters
        -----------
        attachment: int | SlimsAttachment
            The primary key of the attachment or an attachment object
        """
        if isinstance(attachment, SlimsAttachment):
            attachment = attachment.pk

        return self.db.slims_api.get(f"repo/{attachment}")

    @lru_cache(maxsize=None)
    def fetch_pk(self, table: SLIMS_TABLES, *args, **kwargs) -> int | None:
        """SlimsClient.fetch but returns the pk of the first returned record"""
        records = self.fetch(table, *args, **kwargs)
        if len(records) > 0:
            return records[0].pk()
        else:
            return None

    def fetch_user(self, user_name: str):
        """Fetches a user by username"""
        return self.fetch("User", user_userName=user_name)

    def add(self, table: SLIMS_TABLES, data: dict):
        """Add a SLIMS record to a given SLIMS table"""
        record = self.db.add(table, data)
        logger.info(f"SLIMS Add: {table}/{record.pk()}")
        return record

    def update(self, table: SLIMS_TABLES, pk: int, data: dict):
        """Update a SLIMS record"""
        record = self.db.fetch_by_pk(table, pk)
        if record is None:
            raise ValueError(f'No data in SLIMS "{table}" table for pk "{pk}"')
        new_record = record.update(data)
        logger.info(f"SLIMS Update: {table}/{pk}")
        return new_record

    def rest_link(self, table: SLIMS_TABLES, **kwargs):
        """Construct a url link to a SLIMS table with arbitrary filters"""
        base_url = f"{self.url}/rest/{table}"
        queries = [f"?{k}={v}" for k, v in kwargs.items()]
        return base_url + "".join(queries)

    def add_model(
        self, model: SlimsBaseModelTypeVar, *args, **kwargs
    ) -> SlimsBaseModelTypeVar:
        """Given a SlimsBaseModel object, add it to SLIMS
        Args
            model (SlimsBaseModel): object to add
            *args (str): fields to include in the serialization
            **kwargs: passed to model.model_dump()

        Returns
            An instance of the same type of model, with data from
            the resulting SLIMS record
        """
        fields_to_include = set(args) or None
        fields_to_exclude = set(kwargs.get("exclude", []))
        fields_to_exclude.update(["pk", "attachments", "slims_api"])
        rtn = self.add(
            model._slims_table,
            model.model_dump(
                include=fields_to_include,
                exclude=fields_to_exclude,
                **kwargs,
                by_alias=True,
                context="slims_post",
            ),
        )
        return type(model).model_validate(rtn)

    def update_model(self, model: SlimsBaseModel, *args, **kwargs):
        """Given a SlimsBaseModel object, update its (existing) SLIMS record

        Args
            model (SlimsBaseModel): object to update
            *args (str): fields to include in the serialization
            **kwargs: passed to model.model_dump()

        Returns
            An instance of the same type of model, with data from
            the resulting SLIMS record
        """
        if model.pk is None:
            raise ValueError("Cannot update model without a pk")

        fields_to_include = set(args) or None
        rtn = self.update(
            model._slims_table,
            model.pk,
            model.model_dump(
                include=fields_to_include, by_alias=True, **kwargs, context="slims_post"
            ),
        )
        return type(model).model_validate(rtn)

    def add_attachment_content(
        self,
        record: SlimsBaseModel,
        name: str,
        content: bytes | str,
    ) -> int:
        """Add an attachment to a SLIMS record

        Returns
        -------
        int: Primary key of the attachment added.

        Notes
        -----
        - Returned attachment does not contain the name of the attachment in
         Slims, this requires a separate fetch.
        """
        if record.pk is None:
            raise ValueError("Cannot add attachment to a record without a pk")

        if isinstance(content, str):
            content = content.encode("utf-8")

        response = self.db.slims_api.post(
            url="repo",
            body={
                "attm_name": name,
                "atln_recordPk": record.pk,
                "atln_recordTable": record._slims_table,
                "contents": base64.b64encode(content).decode("utf-8"),
            },
        )
        response.raise_for_status()
        return int(response.text)



"""Init operations dir"""

from aind_slims_api.operations.ecephys_session import (
    EcephysSession,
    fetch_ecephys_sessions,
)
from aind_slims_api.operations.histology_procedures import (
    SPIMHistologyExpBlock,
    fetch_histology_procedures,
)

__all__ = [
    "EcephysSession",
    "fetch_ecephys_sessions",
    "SPIMHistologyExpBlock",
    "fetch_histology_procedures",
    "fetch_imaging_metadata"
]



"""Contains models for workflows stored in SLIMS."""

from datetime import datetime
from typing import Annotated, Any, ClassVar, Optional

from pydantic import Field

from aind_slims_api.models.base import SlimsBaseModel
from aind_slims_api.models.utils import UnitSpec


class SlimsExperimentRunStepContent(SlimsBaseModel):
    """Model for a SLIMS ExperimentRunStepContent"""

    pk: Optional[int] = Field(
        default=None, serialization_alias="xrsc_pk", validation_alias="xrsc_pk"
    )
    mouse_pk: Optional[int] = Field(
        default=None,
        serialization_alias="xrsc_fk_content",
        validation_alias="xrsc_fk_content",
    )
    runstep_pk: Optional[int] = Field(
        default=None,
        serialization_alias="xrsc_fk_experimentRunStep",
        validation_alias="xrsc_fk_experimentRunStep",
    )
    created_on: Optional[datetime] = Field(
        default=None,
        serialization_alias="xrsc_createdOn",
        validation_alias="xrsc_createdOn",
    )
    _slims_table = "ExperimentRunStepContent"


class SlimsExperimentRunStep(SlimsBaseModel):
    """Model for a Slims ExperimentRunStep"""

    pk: Optional[int] = Field(
        default=None, serialization_alias="xprs_pk", validation_alias="xprs_pk"
    )
    name: Optional[str] = Field(
        default=None, serialization_alias="xprs_name", validation_alias="xprs_name"
    )
    created_on: Optional[datetime] = Field(
        default=None,
        serialization_alias="xprs_createdOn",
        validation_alias="xprs_createdOn",
    )
    modified_by: Optional[str] = Field(
        default=None,
        serialization_alias="xprs_modifiedBy",
        validation_alias="xprs_modifiedBy",
    )
    experimentrun_pk: Optional[int] = Field(
        default=None,
        serialization_alias="xprs_fk_experimentRun",
        validation_alias="xprs_fk_experimentRun",
    )
    experiment_template_pk: Optional[int] = Field(
        default=None,
        serialization_alias="xpst_fk_experimentTemplate",
        validation_alias="xpst_fk_experimentTemplate",
    )
    _slims_table = "ExperimentRunStep"


class SlimsExperimentTemplate(SlimsBaseModel):
    """Model for a Slims Experiment Template"""

    pk: Optional[int] = Field(
        default=None, serialization_alias="xptm_pk", validation_alias="xptm_pk"
    )
    name: Optional[str] = Field(
        default=None, serialization_alias="xptm_name", validation_alias="xptm_name"
    )
    version_name: Optional[str] = Field(
        default=None,
        serialization_alias="xptm_nameVersion",
        validation_alias="xptm_nameVersion",
    )
    created_on: Optional[datetime] = Field(
        default=None,
        serialization_alias="xptm_createdOn",
        validation_alias="xptm_createdOn",
    )
    _slims_table = "ExperimentTemplate"


class SlimsGroupOfSessionsRunStep(SlimsExperimentRunStep):
    """Model for a Slims ExperimentRunStep"""

    session_type: Optional[str] = Field(
        default=None,
        serialization_alias="xprs_cf_sessionType",
        validation_alias="xprs_cf_sessionType",
    )
    mouse_platform_name: Optional[str] = Field(
        default=None,
        serialization_alias="xprs_cf_mousePlatformName",
        validation_alias="xprs_cf_mousePlatformName",
    )
    active_mouse_platform: Optional[bool] = Field(
        default=None,
        serialization_alias="xprs_cf_activeMousePlatform",
        validation_alias="xprs_cf_activeMousePlatform",
    )
    instrument_pk: Optional[int] = Field(
        default=None,
        serialization_alias="xprs_cf_fk_instrumentJson",
        validation_alias="xprs_cf_fk_instrumentJson",
    )
    # TODO: add device calibrations once we have an example
    # device_calibrations_attachment: Optional[str] = Field(
    #     default=None,
    #     serialization_alias="xprs_cf_deviceCalibrations",
    #     validation_alias="xprs_cf_deviceCalibrations"
    # )
    _base_fetch_filters: ClassVar[dict[str, str]] = {
        "xprs_name": "Group of Sessions",
    }


class SlimsMouseSessionRunStep(SlimsExperimentRunStep):
    """Model for a Slims ExperimentRunStep"""

    _base_fetch_filters: ClassVar[dict[str, str]] = {
        "xprs_name": "Mouse Session",
    }


class SlimsProtocolRunStep(SlimsExperimentRunStep):
    """Model for a Slims Protocol ExperimentRunStep"""

    protocol_pk: Optional[int] = Field(
        None,
        serialization_alias="xprs_cf_fk_protocol",
        validation_alias="xprs_cf_fk_protocol",
    )
    _base_fetch_filters: ClassVar[dict[str, str]] = {
        "xprs_name": "Protocol",
    }


class SlimsWashRunStep(SlimsExperimentRunStep):
    """Model for a Slims Wash ExperimentRunStep"""

    wash_name: Optional[str] = Field(
        None, serialization_alias="xprs_name", validation_alias="xprs_name"
    )
    spim_wash_type: Optional[str] = Field(
        None,
        serialization_alias="xprs_cf_spimWashType",
        validation_alias="xprs_cf_spimWashType",
    )  # FIXED CHOICE
    reagent_pk: Optional[list[int]] = Field(
        None,
        serialization_alias="xprs_cf_fk_reagent_multiselect",
        validation_alias="xprs_cf_fk_reagent_multiselect",
    )
    start_time: Optional[datetime] = Field(
        None,
        serialization_alias="xprs_cf_sbipDelipidationWash1Start",
        validation_alias="xprs_cf_sbipDelipidationWash1Start",
    )
    end_time: Optional[datetime] = Field(
        None,
        serialization_alias="xprs_cf_sbipDelipidationWash6End",
        validation_alias="xprs_cf_sbipDelipidationWash6End",
    )
    mass: Annotated[float | None, UnitSpec("&mu;g")] = Field(
        None, serialization_alias="xprs_cf_mass", validation_alias="xprs_cf_mass"
    )
    notes: Optional[str] = Field(
        None, serialization_alias="xprs_cf_notes", validation_alias="xprs_cf_notes"
    )
    _base_fetch_filters: ClassVar[dict[str, Any]] = {
        "xprs_name": [
            "Wash 1",
            "Wash 2",
            "Wash 3",
            "Wash 4",
            "Refractive Index Matching Wash",
            "Primary Antibody Wash",
            "Secondary Antibody Wash",
            "MBS Wash",
            "Gelation PBS Wash",
            "Stock X + VA-044 Equilibration",
            "Gelation + ProK RT",
            "Gelation + Add'l ProK 37C",
            "Final PBS Wash",
        ]
    }


class SlimsSPIMImagingRunStep(SlimsExperimentRunStep):
    """Model for a Slims SPIM Imaging ExperimentRunStep"""

    # protocol, imaging metadata
    _base_fetch_filters: ClassVar[dict[str, str]] = {
        "xprs_name": "Imaging Metadata",
    }
